/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "operator/operator.pmh"

#define FE_PPO_DEBUG		FALSE

using namespace fe;

void PuppetOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	//* TODO
	catalogRemove("IdAttr");

	catalog<bool>("dampJoints")=false;
	catalog<String>("dampJoints","label")="Restrain";
	catalog<String>("dampJoints","page")="Edit";
	catalog<bool>("dampJoints","joined")=true;

	catalog<Real>("dampRatio")=1.0;
	catalog<String>("dampRatio","label")="Freedom";
	catalog<String>("dampRatio","enabler")="dampJoints";
	catalog<String>("dampRatio","page")="Edit";

	//* pose is alternative reference
	catalog<bool>("poseJoints")=false;
	catalog<String>("poseJoints","label")="Rest Repose";
	catalog<String>("poseJoints","page")="Edit";
	catalog<bool>("poseJoints","joined")=true;

	//* how much the pose changes the reference
	catalog<Real>("poseRatio")=1.0;
	catalog<String>("poseRatio","label")="Ratio";
	catalog<String>("poseRatio","enabler")="poseJoints";
	catalog<String>("poseRatio","page")="Edit";

	catalog<bool>("offsetJoints")=false;
	catalog<String>("offsetJoints","label")="Offset";
	catalog<String>("offsetJoints","page")="Edit";
	catalog<bool>("offsetJoints","joined")=true;

	catalog<Real>("offsetRatio")=1.0;
	catalog<String>("offsetRatio","label")="Scale";
	catalog<String>("offsetRatio","enabler")="offsetJoints";
	catalog<String>("offsetRatio","page")="Edit";

	catalog<bool>("drawNames")=false;
	catalog<String>("drawNames","label")="Draw Names";
	catalog<String>("drawNames","page")="Display";

	catalog<bool>("recolor")=false;
	catalog<String>("recolor","label")="Uniform Color";
	catalog<bool>("recolor","joined")=true;
	catalog<String>("recolor","page")="Display";

	catalog<SpatialVector>("jointColor")=SpatialVector(0.0,1.0,0.0);
	catalog<String>("jointColor","label")="As";
	catalog<String>("jointColor","page")="Display";

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","visible")="any";
	catalog<String>("Brush","prompt")="prompt here";

	catalog< sp<Component> >("Input Joints");

	catalog< sp<Component> >("Pose Joints");
	catalog<bool>("Pose Joints","optional")=true;

	catalog< sp<Component> >("Offset Joints");
	catalog<bool>("Offset Joints","optional")=true;

	m_spWireframe=new DrawMode();
	m_spWireframe->setDrawStyle(DrawMode::e_wireframe);

	m_spThick=new DrawMode();
	m_spThick->copy(m_spWireframe);
	m_spThick->setLineWidth(2.0);

	m_spSolid=new DrawMode();
	m_spSolid->copy(m_spWireframe);
	m_spSolid->setDrawStyle(DrawMode::e_foreshadow);
	m_spSolid->setLit(FALSE);
	m_spSolid->setBackfaceCulling(FALSE);

	m_spOverlay=new DrawMode();
	m_spOverlay->copy(m_spWireframe);
	m_spOverlay->setDrawStyle(DrawMode::e_solid);
	m_spOverlay->setLit(FALSE);
	m_spOverlay->setBackfaceCulling(FALSE);

	m_spManipulator=registry()->create("*.ValueManipulator");
	if(m_spManipulator.isValid())
	{
		sp<Catalog> spCatalog=m_spManipulator;
		if(spCatalog.isNull())
		{
			m_spManipulator=NULL;
		}
		else
		{
			const Color red(1.0,0.0,0.0);
			const Color green(0.0,1.0,0.0);
			const Color blue(0.0,0.0,1.0);
			const Color grey(0.5,0.5,0.5);

			addParam("freedom","Freedom",1.0,0.0,1.0,grey);
			addParam("freedomX","Freedom X",1.0,0.0,1.0,red);
			addParam("freedomY","Freedom Y",1.0,0.0,1.0,green);
			addParam("freedomZ","Freedom Z",1.0,0.0,1.0,blue);

			addParam("minX","Min X",180.0,0.0,180.0,red);
			addParam("maxX","Max X",180.0,0.0,180.0,red);

			addParam("minY","Min Y",180.0,0.0,180.0,green);
			addParam("maxY","Max Y",180.0,0.0,180.0,green);

			addParam("minZ","Min Z",180.0,0.0,180.0,blue);
			addParam("maxZ","Max Z",180.0,0.0,180.0,blue);

			spCatalog->catalog<bool>("gripLabels")=TRUE;
			spCatalog->catalog<Real>("dimming")=0.5;
		}
	}
}

void PuppetOp::addParam(String a_key,String a_label,Real a_value,
	Real a_min,Real a_max,Color a_color)
{
	sp<Catalog> spCatalog=m_spManipulator;
	FEASSERT(spCatalog.isValid());

	spCatalog->catalog<Real>(a_key)=a_value;
	spCatalog->catalog<Real>(a_key,"min")=a_min;
	spCatalog->catalog<Real>(a_key,"max")=a_max;
	spCatalog->catalog<String>(a_key,"label")=a_label;
	spCatalog->catalog<SpatialVector>(a_key,"color")=a_color;
}

void PuppetOp::handle(Record& a_rSignal)
{
#if FE_PPO_DEBUG
	feLog("PuppetOp::handle\n");
#endif

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Joints")) return;

	sp<SurfaceAccessorI> spInputName;
	if(!access(spInputName,spInputAccessible,e_primitive,"name")) return;

	sp<SurfaceAccessorI> spInputParent;
	if(!access(spInputParent,spInputAccessible,e_primitive,"parent")) return;

	sp<SurfaceAccessorI> spInputColor;
	if(!access(spInputColor,spInputAccessible,e_point,"Cd")) return;

	const U32 primitiveCount=spInputName->count();

	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

	sp<DrawI> spDrawOverlay;
	accessBrushOverlay(spDrawOverlay,a_rSignal,e_quiet);

#if FE_PPO_DEBUG
	feLog("PuppetOp::handle brush %d brushed %d\n",
			spDrawBrush.isValid(),m_brushed);
#endif

	if(spDrawBrush.isValid())
	{
		sp<SurfaceAccessorI> spOutputVertices;
		if(m_spOutputAccessible.isNull() ||
				!access(spOutputVertices,m_spOutputAccessible,
				e_primitive,e_vertices)) return;

		spDrawBrush->pushDrawMode(m_spWireframe);
		spDrawOverlay->pushDrawMode(m_spOverlay);

		sp<ViewI> spView=spDrawOverlay->view();

		m_brushed=TRUE;

		const Color black(0.0,0.0,0.0);
		const Color white(1.0,1.0,0.8);
		const Color grey(0.8,0.8,1.0);
		const Color red(0.8,0.0,0.0);
		const Color green(0.0,0.5,0.0);
		const Color blue(0.0,0.0,1.0);
		const Color cyan(0.0,1.0,1.0);
		const Color magenta(0.8,0.0,0.8);
		const Color yellow(1.0,1.0,0.0);
		const Color sparkle(1.0,1.0,0.6);
		const Color orange(1.0,0.5,0.0);
		const Color blank(0.0,0.0,0.0,0.0);

		m_event.bind(windowEvent(a_rSignal));

#if FE_PPO_DEBUG
		feLog("%s\n",c_print(m_event));
#endif

		if(m_event.isKeyPress(WindowEvent::e_keyEscape))
		{
			m_pickName="";
		}

		if(m_event.isKeyPress(WindowEvent::e_keyBackspace))
		{
			deleteJoint(m_pickName);
			m_pickName="";
		}

		const Vector2i mouse(m_event.mouseX(),m_event.mouseY());

		std::map< String, std::vector<String> > childMap;
		for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
		{
			const String jointName=spInputName->string(primitiveIndex);
			const String parentName=spInputParent->string(primitiveIndex);

			childMap[parentName].push_back(jointName);
		}

		const Real nearMax=64.0;
		const Real nearMax2=nearMax*nearMax;

		Real nearDist2= -1.0;
		String nearName;

		for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;
				primitiveIndex++)
		{
			const String jointName=spInputName->string(primitiveIndex);

			std::vector<String>& rChildList=childMap[jointName];
			const U32 childCount=rChildList.size();

			const String otherName=childCount? rChildList[0]: jointName;

			const I32 subCount=spOutputVertices->subCount(primitiveIndex);
			if(subCount<3)
			{
				continue;
			}

			const SpatialTransform& xform=m_animMap[jointName];
			const SpatialTransform& xformOther=m_animMap[otherName];

			const SpatialVector midpoint=
					0.5*(xform.translation()+xformOther.translation());

			const Vector2i projected=
					spView->project(midpoint,ViewMode::e_perspective);

			const Real dist2=magnitudeSquared(projected-mouse);
			if(dist2>nearMax2)
			{
				continue;
			}

			if(nearDist2<0.0 || nearDist2>dist2)
			{
				nearDist2=dist2;
				nearName=jointName;
			}
		}

		if(!m_dragging && m_event.isMouseRelease(WindowEvent::e_itemLeft))
		{
			m_pickName=nearName;

			if(m_spManipulator.isValid())
			{
				sp<Catalog> spCatalog=m_spManipulator;
				FEASSERT(spCatalog.isValid());

				Record jointRecord=lookupJoint(m_pickName);

				spCatalog->catalog<String>("selection")=m_pickName;
				spCatalog->catalog<bool>("show")=true;

				spCatalog->catalog<Real>("freedom")=m_aFreedom(jointRecord);
				spCatalog->catalog<Real>("freedomX")=m_aFreedomX(jointRecord);
				spCatalog->catalog<Real>("freedomY")=m_aFreedomY(jointRecord);
				spCatalog->catalog<Real>("freedomZ")=m_aFreedomZ(jointRecord);

				spCatalog->catalog<Real>("minX")=m_aMinX(jointRecord);
				spCatalog->catalog<Real>("minY")=m_aMinY(jointRecord);
				spCatalog->catalog<Real>("minZ")=m_aMinZ(jointRecord);

				spCatalog->catalog<Real>("maxX")=m_aMaxX(jointRecord);
				spCatalog->catalog<Real>("maxY")=m_aMaxY(jointRecord);
				spCatalog->catalog<Real>("maxZ")=m_aMaxZ(jointRecord);
			}
		}

		m_dragging=(m_event.isMouseDrag() && m_event.mouseButtons());

		const bool recolor=catalog<bool>("recolor");
		const SpatialVector jointColor=catalog<SpatialVector>("jointColor");

		SpatialVector line[4];

		const U32 resolution=16;

		SpatialVector pointArray[resolution];
		SpatialVector normalArray[resolution];
		Real radiusArray[resolution];
		Color colorArray[resolution];

		for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
		{
			const I32 subCount=spOutputVertices->subCount(primitiveIndex);
			if(subCount<3)
			{
				continue;
			}

			const String jointName=spInputName->string(primitiveIndex);

			const I32 pointIndex0=
					spOutputVertices->integer(primitiveIndex,0);
			const I32 pointIndex1=
					spOutputVertices->integer(primitiveIndex,1);
			const I32 pointIndex2=
					spOutputVertices->integer(primitiveIndex,2);

			const BWORD bright=(jointName==m_pickName) ||
					(!m_dragging && jointName==nearName);

			const Color color0=makeColor(recolor? jointColor:
					spInputColor->spatialVector(pointIndex0),bright);

			const Color color1=
					makeColor(spInputColor->spatialVector(pointIndex1),bright);
			const Color color2=
					makeColor(spInputColor->spatialVector(pointIndex2),bright);

			const SpatialVector point0=
					spOutputVertices->spatialVector(primitiveIndex,0);
			const SpatialVector point1=
					spOutputVertices->spatialVector(primitiveIndex,1);
			const SpatialVector point2=
					spOutputVertices->spatialVector(primitiveIndex,2);

			const SpatialVector mod1=2.0*point0-point1;
			const SpatialVector mod2=2.0*point0-point2;

			SpatialVector midpoint=point0;
			SpatialVector endpoint=point0;

			std::vector<String>& rChildList=childMap[jointName];
			const U32 childCount=rChildList.size();
			for(U32 childIndex=0;childIndex<childCount;childIndex++)
			{
				const String childName=rChildList[childIndex];

				const SpatialTransform& xformChild=m_animMap[childName];

				const SpatialVector& childPoint0=xformChild.translation();

				line[0]=childPoint0;

				line[1]=point1;
				spDrawBrush->drawLines(line,NULL,2,DrawI::e_strip,
						false,&color1);

				line[1]=point2;
				spDrawBrush->drawLines(line,NULL,2,DrawI::e_strip,
						false,&color2);

				line[1]=mod1;
				spDrawBrush->drawLines(line,NULL,2,DrawI::e_strip,
						false,&color0);

				line[1]=mod2;
				spDrawBrush->drawLines(line,NULL,2,DrawI::e_strip,
						false,&color0);

				if(!childIndex)
				{
					midpoint=0.5*(childPoint0+point0);
					endpoint=childPoint0;
				}
			}

			line[0]=point1;
			line[1]=mod2;
			line[2]=mod1;
			line[3]=point2;
			spDrawBrush->drawLines(line,NULL,4,DrawI::e_strip,
					false,&color0);

			const BWORD drawNames=catalog<bool>("drawNames");

			if(drawNames)
			{
				const U32 margin=1;
				const Vector2i projected=
						spView->project(midpoint,ViewMode::e_perspective);
				drawLabel(spDrawOverlay,projected,"  "+jointName,
						FALSE,margin,white,NULL,&blue);
			}

			line[0]=point0;
			line[1]=endpoint;

			spDrawBrush->pushDrawMode(m_spThick);

			const Real dotRadius=6;
			if(jointName==m_pickName)
			{
				drawDot(spView,spDrawOverlay,midpoint,dotRadius,yellow);

				spDrawBrush->drawLines(line,NULL,2,DrawI::e_strip,
						false,&yellow);

				spDrawBrush->pushDrawMode(m_spSolid);

				Record jointRecord=lookupJoint(m_pickName);

				const SpatialTransform& xform=m_animMap[jointName];

				for(U32 axis=0;axis<3;axis++)
				{
					const Real jointFreedom=lookupFreedom(jointRecord,axis);
					const Real jointMin= -lookupMin(jointRecord,axis)*degToRad;
					const Real jointMax=lookupMax(jointRecord,axis)*degToRad;

					for(U32 pass=0;pass<2;pass++)
					{
						const Real freescale=pass? jointFreedom: 1.0;

						for(U32 m=0;m<resolution;m++)
						{
							const Real fraction=m/(resolution-1.0);

							const Real start=jointMin*freescale;
							const Real length=(jointMax-jointMin)*freescale;

							SpatialTransform spun=xform;
							rotate(spun,(start+length*fraction),Axis(axis));

							set(pointArray[m]);
							pointArray[m][(axis==1)? 0: 1]=4.0;
							transformVector(spun,pointArray[m],pointArray[m]);

							set(normalArray[m]);
							normalArray[m][axis]=1.0;
							transformVector(spun,normalArray[m],normalArray[m]);

							radiusArray[m]=pass? 0.2: 0.1;

							set(colorArray[m]);
							colorArray[m][axis]=pass? 1.0: 0.5;
						}

						const BWORD multicolor=TRUE;
						const OperateCommon::CurveMode curveMode=
								OperateCommon::e_tube;
						const SpatialVector towardCamera(0.0,0.0,1.0);

						drawTube(spDrawBrush,pointArray,normalArray,radiusArray,
								resolution,curveMode,multicolor,colorArray,
								towardCamera);
					}
				}

				spDrawBrush->popDrawMode();
			}
			else if(!m_dragging && jointName==nearName)
			{
				drawDot(spView,spDrawOverlay,midpoint,dotRadius,cyan);

				spDrawBrush->drawLines(line,NULL,2,DrawI::e_strip,
						false,&cyan);
			}
			else if(drawNames)
			{
				drawDot(spView,spDrawOverlay,midpoint,dotRadius,magenta);
			}

			spDrawBrush->popDrawMode();
		}

		if(m_pickName.empty())
		{
			if(m_spManipulator.isValid())
			{
				sp<Catalog> spCatalog=m_spManipulator;
				FEASSERT(spCatalog.isValid());

				spCatalog->catalog<String>("selection")="";
				spCatalog->catalog<bool>("show")=false;
			}
		}
		else if(m_spManipulator.isValid())
		{
			sp<Catalog> spCatalog=m_spManipulator;
			FEASSERT(spCatalog.isValid());

			Record jointRecord=lookupJoint(m_pickName);

			m_aFreedom(jointRecord)=spCatalog->catalog<Real>("freedom");
			m_aFreedomX(jointRecord)=spCatalog->catalog<Real>("freedomX");
			m_aFreedomY(jointRecord)=spCatalog->catalog<Real>("freedomY");
			m_aFreedomZ(jointRecord)=spCatalog->catalog<Real>("freedomZ");

			m_aMinX(jointRecord)=spCatalog->catalog<Real>("minX");
			m_aMinY(jointRecord)=spCatalog->catalog<Real>("minY");
			m_aMinZ(jointRecord)=spCatalog->catalog<Real>("minZ");

			m_aMaxX(jointRecord)=spCatalog->catalog<Real>("maxX");
			m_aMaxY(jointRecord)=spCatalog->catalog<Real>("maxY");
			m_aMaxZ(jointRecord)=spCatalog->catalog<Real>("maxZ");
		}

		if(m_spManipulator.isValid())
		{
			m_spManipulator->bindOverlay(spDrawOverlay);
			m_spManipulator->handle(a_rSignal);
			m_spManipulator->draw(spDrawBrush,NULL);
		}

		spDrawOverlay->popDrawMode();
		spDrawBrush->popDrawMode();

		return;
	}

	const BWORD dampJoints=catalog<bool>("dampJoints");
	const Real dampRatio=dampJoints? catalog<Real>("dampRatio"): 1.0;
	const Real poseRatio=
			catalog<bool>("poseJoints")? catalog<Real>("poseRatio"): 0.0;
	const Real offsetRatio=
			catalog<bool>("offsetJoints")? catalog<Real>("offsetRatio"): 0.0;

	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,m_spOutputAccessible,e_point,e_position)) return;

	sp<SurfaceAccessorI> spOutputAnimX;
	if(!access(spOutputAnimX,m_spOutputAccessible,e_primitive,"animX")) return;

	sp<SurfaceAccessorI> spOutputAnimY;
	if(!access(spOutputAnimY,m_spOutputAccessible,e_primitive,"animY")) return;

	sp<SurfaceAccessorI> spOutputAnimZ;
	if(!access(spOutputAnimZ,m_spOutputAccessible,e_primitive,"animZ")) return;

	sp<SurfaceAccessorI> spOutputAnimT;
	if(!access(spOutputAnimT,m_spOutputAccessible,e_primitive,"animT")) return;

	sp<SurfaceAccessorI> spInputRefX;
	sp<SurfaceAccessorI> spInputRefY;
	sp<SurfaceAccessorI> spInputRefZ;
	sp<SurfaceAccessorI> spInputRefT;
	if(poseRatio<1.0)
	{
		if(!access(spInputRefX,spInputAccessible,e_primitive,"refX")) return;
		if(!access(spInputRefY,spInputAccessible,e_primitive,"refY")) return;
		if(!access(spInputRefZ,spInputAccessible,e_primitive,"refZ")) return;
		if(!access(spInputRefT,spInputAccessible,e_primitive,"refT")) return;
	}

	m_posePrimitiveMap.clear();

	sp<SurfaceAccessibleI> spPoseAccessible;
	sp<SurfaceAccessorI> spPoseAnimX;
	sp<SurfaceAccessorI> spPoseAnimY;
	sp<SurfaceAccessorI> spPoseAnimZ;
	sp<SurfaceAccessorI> spPoseAnimT;
	if(poseRatio>0.0)
	{
		access(spPoseAccessible,"Pose Joints",e_quiet);

		if(spPoseAccessible.isValid())
		{
			sp<SurfaceAccessorI> spPoseName;
			if(!access(spPoseName,spPoseAccessible,
					e_primitive,"name")) return;

			if(!access(spPoseAnimX,spPoseAccessible,
					e_primitive,"animX")) return;

			if(!access(spPoseAnimY,spPoseAccessible,
					e_primitive,"animY")) return;

			if(!access(spPoseAnimZ,spPoseAccessible,
					e_primitive,"animZ")) return;

			if(!access(spPoseAnimT,spPoseAccessible,
					e_primitive,"animT")) return;

			const U32 posePrimitiveCount=spPoseName->count();
			for(U32 posePrimitiveIndex=0;posePrimitiveIndex<posePrimitiveCount;
					posePrimitiveIndex++)
			{
				const String jointName=spPoseName->string(posePrimitiveIndex);

				m_posePrimitiveMap[jointName]=posePrimitiveIndex;
			}
		}
		else
		{
			catalog<String>("warning")+="no pose input;";
		}
	}

	m_offsetPrimitiveMap.clear();

	sp<SurfaceAccessibleI> spOffsetAccessible;
	sp<SurfaceAccessorI> spOffsetAnimX;
	sp<SurfaceAccessorI> spOffsetAnimY;
	sp<SurfaceAccessorI> spOffsetAnimZ;
	sp<SurfaceAccessorI> spOffsetAnimT;
	if(offsetRatio>0.0)
	{
		access(spOffsetAccessible,"Offset Joints",e_quiet);

		if(spOffsetAccessible.isValid())
		{
			sp<SurfaceAccessorI> spOffsetName;
			if(!access(spOffsetName,spOffsetAccessible,
					e_primitive,"name")) return;

			if(!access(spOffsetAnimX,spOffsetAccessible,
					e_primitive,"animX")) return;

			if(!access(spOffsetAnimY,spOffsetAccessible,
					e_primitive,"animY")) return;

			if(!access(spOffsetAnimZ,spOffsetAccessible,
					e_primitive,"animZ")) return;

			if(!access(spOffsetAnimT,spOffsetAccessible,
					e_primitive,"animT")) return;

			const U32 offsetPrimitiveCount=spOffsetName->count();
			for(U32 offsetPrimitiveIndex=0;
					offsetPrimitiveIndex<offsetPrimitiveCount;
					offsetPrimitiveIndex++)
			{
				const String jointName=
						spOffsetName->string(offsetPrimitiveIndex);

				m_offsetPrimitiveMap[jointName]=offsetPrimitiveIndex;
			}
		}
		else
		{
			catalog<String>("warning")+="no offset input;";
		}
	}

	m_animMap.clear();

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const String jointName=spInputName->string(primitiveIndex);

		Record jointRecord=lookupJoint(jointName);

		SpatialVector oneDamp(1.0,1.0,1.0);

		if(dampJoints)
		{
			const SpatialVector jointFreedom(
					lookupFreedom(jointRecord,0),
					lookupFreedom(jointRecord,1),
					lookupFreedom(jointRecord,2));

			oneDamp=jointFreedom*dampRatio;
		}

		const SpatialVector jointMin(
				-lookupMin(jointRecord,0)*degToRad,
				-lookupMin(jointRecord,1)*degToRad,
				-lookupMin(jointRecord,2)*degToRad);

		const SpatialVector jointMax(
				lookupMax(jointRecord,0)*degToRad,
				lookupMax(jointRecord,1)*degToRad,
				lookupMax(jointRecord,2)*degToRad);

		SpatialTransform xformLocalRef;
		SpatialQuaternion quatLocalRef;
		if(spInputRefX.isValid())
		{
			xformLocalRef.column(0)=spInputRefX->spatialVector(primitiveIndex);
			xformLocalRef.column(1)=spInputRefY->spatialVector(primitiveIndex);
			xformLocalRef.column(2)=spInputRefZ->spatialVector(primitiveIndex);
			xformLocalRef.column(3)=spInputRefT->spatialVector(primitiveIndex);

			quatLocalRef=xformLocalRef;
		}
		else
		{
			setIdentity(xformLocalRef);
			set(quatLocalRef);
		}

		SpatialTransform xformLocal;
		xformLocal.column(0)=spOutputAnimX->spatialVector(primitiveIndex);
		xformLocal.column(1)=spOutputAnimY->spatialVector(primitiveIndex);
		xformLocal.column(2)=spOutputAnimZ->spatialVector(primitiveIndex);
		xformLocal.column(3)=spOutputAnimT->spatialVector(primitiveIndex);

		if(spPoseAccessible.isValid())
		{
			const U32 posePrimitiveIndex=m_posePrimitiveMap[jointName];

			SpatialTransform xformPose;
			xformPose.column(0)=spPoseAnimX->spatialVector(posePrimitiveIndex);
			xformPose.column(1)=spPoseAnimY->spatialVector(posePrimitiveIndex);
			xformPose.column(2)=spPoseAnimZ->spatialVector(posePrimitiveIndex);
			xformPose.column(3)=spPoseAnimT->spatialVector(posePrimitiveIndex);

			if(poseRatio<1.0)
			{
				SpatialTransform invLocalRef;
				invert(invLocalRef,xformLocalRef);

				const SpatialQuaternion quatToPose=xformPose*invLocalRef;

				Real radians;
				SpatialVector axis;
				quatToPose.computeAngleAxis(radians,axis);
				const SpatialQuaternion quatToPartial(poseRatio*radians,axis);

				SpatialTransform xformToPartial=SpatialTransform(quatToPartial);
				xformToPartial.translation();

				xformLocalRef=xformToPartial*xformLocalRef;
			}
			else
			{
				xformLocalRef=xformPose;
			}

			quatLocalRef=xformLocalRef;
		}

		SpatialQuaternion quatOffset(0.0,0.0,0.0,1.0);
		if(spOffsetAccessible.isValid())
		{
			const U32 offsetPrimitiveIndex=m_offsetPrimitiveMap[jointName];

			SpatialTransform xformOffset;
			xformOffset.column(0)=
					spOffsetAnimX->spatialVector(offsetPrimitiveIndex);
			xformOffset.column(1)=
					spOffsetAnimY->spatialVector(offsetPrimitiveIndex);
			xformOffset.column(2)=
					spOffsetAnimZ->spatialVector(offsetPrimitiveIndex);
			xformOffset.column(3)=
					spOffsetAnimT->spatialVector(offsetPrimitiveIndex);

			SpatialTransform invLocalRef;
			invert(invLocalRef,xformLocalRef);

			quatOffset=xformOffset*invLocalRef;
		}

		SpatialQuaternion quatLocal=xformLocal;

		//* min max
		const BWORD constrained=
				(jointMin[0]<180.0 || jointMin[1]<180.0 || jointMin[2]<180.0 ||
				jointMax[0]<180.0 || jointMax[1]<180.0 || jointMax[2]<180.0);

		SpatialVector displaceBlended;
		SpatialQuaternion quatBlended;
		if(constrained || oneDamp[0]<1.0 || oneDamp[1]<1.0 || oneDamp[2]<1.0)
		{
			if(!constrained &&
					oneDamp[0] == oneDamp[1] && oneDamp[1] == oneDamp[2])
			{
				slerp(quatBlended,quatLocalRef,quatLocal,oneDamp[0]);

				displaceBlended=oneDamp[0]*xformLocal.translation()+
						(1.0-oneDamp[0])*xformLocalRef.translation();
			}
			else
			{
				const SpatialQuaternion quatDelta=(-quatLocalRef)*quatLocal;

				SpatialEuler euler=SpatialTransform(quatDelta);
				for(U32 m=0;m<3;m++)
				{
					euler[m]*=oneDamp[m];

					if(euler[m]<jointMin[m])
					{
						euler[m]=jointMin[m];
					}
					if(euler[m]>jointMax[m])
					{
						euler[m]=jointMax[m];
					}
					displaceBlended[m]=oneDamp[m]*xformLocal.translation()[m]+
							(1.0-oneDamp[m])*xformLocalRef.translation()[m];
				}

				const SpatialQuaternion quatPartial=SpatialTransform(euler);

				quatBlended=quatLocalRef*quatPartial;
			}
		}
		else
		{
			quatBlended=quatLocal;
			displaceBlended=xformLocal.translation();
		}

		if(offsetRatio>0.0)
		{
			Real radiansOff;
			SpatialVector axisOff;
			quatOffset.computeAngleAxis(radiansOff,axisOff);
			const SpatialQuaternion quatPartialOff(
					offsetRatio*radiansOff,axisOff);

			quatBlended=quatBlended*quatPartialOff;
		}

		SpatialTransform xformBlended(quatBlended);
		xformBlended.translation()=displaceBlended;

		spOutputAnimX->set(primitiveIndex,xformBlended.column(0));
		spOutputAnimY->set(primitiveIndex,xformBlended.column(1));
		spOutputAnimZ->set(primitiveIndex,xformBlended.column(2));
		spOutputAnimT->set(primitiveIndex,xformBlended.column(3));

		//* update visible triangle

		const String parentName=spInputParent->string(primitiveIndex);

		SpatialTransform xformParent;

		const BWORD hasParent=(m_animMap.find(parentName)==m_animMap.end());
		if(hasParent)
		{
			setIdentity(xformParent);
		}
		else
		{
			xformParent=m_animMap[parentName];
		}

		const SpatialTransform xformWorld=xformBlended*xformParent;
		m_animMap[jointName]=xformWorld;

		//* TODO param
		const Real triScale=1.0;

		const I32 pointIndex0=spOutputVertices->integer(primitiveIndex,0);
		const I32 pointIndex1=spOutputVertices->integer(primitiveIndex,1);
		const I32 pointIndex2=spOutputVertices->integer(primitiveIndex,2);

		const SpatialVector point=xformWorld.translation();
		spOutputPoint->set(pointIndex0,point);
		spOutputPoint->set(pointIndex1,point+triScale*xformWorld.column(0));
		spOutputPoint->set(pointIndex2,point+triScale*xformWorld.column(2));
	}

#if FE_PPO_DEBUG
	feLog("PuppetOp::handle done\n");
#endif
}

Color PuppetOp::makeColor(Color a_color,BWORD a_bright)
{
	if(!a_bright)
	{
		return a_color;
	}

	Color result=a_color*2.0;
	for(U32 m=0;m<3;m++)
	{
		if(result[m]>1.0)
		{
			result[m]=1.0;
		}
		if(result[m]<0.5)
		{
			result[m]=0.5;
		}
	}
	result[3]=a_color[3];

	return result;
}

Record PuppetOp::lookupJoint(String a_jointName)
{
	Record jointRecord=m_spJointMap->lookup(a_jointName);
	if(jointRecord.isValid())
	{
		return jointRecord;
	}

	//* create new one
	jointRecord=m_spScope->createRecord("Joint");

	m_aName(jointRecord)=a_jointName;
	m_aFreedom(jointRecord)=1.0;
	m_aFreedomX(jointRecord)=1.0;
	m_aFreedomY(jointRecord)=1.0;
	m_aFreedomZ(jointRecord)=1.0;

	m_aMinX(jointRecord)=180.0;
	m_aMinY(jointRecord)=180.0;
	m_aMinZ(jointRecord)=180.0;

	m_aMaxX(jointRecord)=180.0;
	m_aMaxY(jointRecord)=180.0;
	m_aMaxZ(jointRecord)=180.0;

	m_spState->add(jointRecord);

	return jointRecord;
}

void PuppetOp::deleteJoint(String a_jointName)
{
	if(a_jointName.empty())
	{
		return;
	}

	Record jointRecord=m_spJointMap->lookup(a_jointName);
	if(jointRecord.isValid())
	{
		m_spState->remove(jointRecord);
	}
}

Real PuppetOp::lookupFreedom(Record a_record,U32 a_axis)
{
	const Real freedom=m_aFreedom(a_record);
	Real freedomAxis=1.0;

	switch(a_axis)
	{
		case 0:
			freedomAxis=m_aFreedomX(a_record);
			break;
		case 1:
			freedomAxis=m_aFreedomY(a_record);
			break;
		case 2:
			freedomAxis=m_aFreedomZ(a_record);
			break;
		default:
			;
	}

	return freedom*freedomAxis;
}

Real PuppetOp::lookupMin(Record a_record,U32 a_axis)
{
	switch(a_axis)
	{
		case 0:
			return m_aMinX(a_record);
		case 1:
			return m_aMinY(a_record);
		case 2:
			return m_aMinZ(a_record);
		default:
			;
	}

	return 0.0;
}

Real PuppetOp::lookupMax(Record a_record,U32 a_axis)
{
	switch(a_axis)
	{
		case 0:
			return m_aMaxX(a_record);
		case 1:
			return m_aMaxY(a_record);
		case 2:
			return m_aMaxZ(a_record);
		default:
			;
	}

	return 0.0;
}

void PuppetOp::setupState(void)
{
	//* esc: clear selection?
	//* return: toggle displace/rotate
	//* delete: reset joint
	//* left: freedom
	//* middle: max

	//* 3x displacement max
	//* 3x rotation max

	//* 3x+1 displacement freedom
	//* 3x+1 rotation freedom

	m_aName.populate(m_spScope,"Joint","Name");

	m_aFreedom.populate(m_spScope,"Joint","Freedom");
	m_aFreedomX.populate(m_spScope,"Joint","FreedomX");
	m_aFreedomY.populate(m_spScope,"Joint","FreedomY");
	m_aFreedomZ.populate(m_spScope,"Joint","FreedomZ");

	m_aMinX.populate(m_spScope,"Joint","MinX");
	m_aMinY.populate(m_spScope,"Joint","MinY");
	m_aMinZ.populate(m_spScope,"Joint","MinZ");

	m_aMaxX.populate(m_spScope,"Joint","MaxX");
	m_aMaxY.populate(m_spScope,"Joint","MaxY");
	m_aMaxZ.populate(m_spScope,"Joint","MaxZ");

	m_spJointMap=m_spState->lookupMap(m_aName);
}

BWORD PuppetOp::loadState(const String& a_rBuffer)
{
	if(OperatorState::loadState(a_rBuffer))
	{
		m_spJointMap=m_spState->lookupMap(m_aName);
		return TRUE;
	}

	return FALSE;
}
